/* $Id: pthread_shared_mtx1.c,v 1.3 2015/11/02 18:45:28 kostik Exp kostik $ */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdatomic.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct iter {
	unsigned iter1;
	unsigned pad1[15];
	unsigned iter2;
	unsigned pad2[15];
	atomic_uint aiter1;
	unsigned pad3[15];
	atomic_uint aiter2;
	unsigned pad4[15];
};

static unsigned maxiter = 10000000;

static void
test_mutex(pthread_mutex_t *m, unsigned *iter, atomic_uint *aiter)
{
	unsigned i;

	for (i = 0; i < maxiter; i++) {
		pthread_mutex_lock(m);
		(*iter)++;
		(void)atomic_fetch_add(aiter, 1);
		pthread_mutex_unlock(m);
	}
}

int
main(int argc, char *argv[])
{
	void *p;
	struct iter *iter;
	pthread_mutexattr_t ma;
	pthread_mutex_t *m;
	pid_t child1, child2, pid;
	int error, i;

	if (argc == 2) {
		maxiter = atoi(argv[1]);
	} else if (argc != 1) {
		fprintf(stderr, "usage: prog [count]\n");
		return (2);
	}

	p = mmap(NULL, sizeof(*iter) + sizeof(*m),
	    PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANON, -1, 0);
	if (p == MAP_FAILED) {
		fprintf(stderr, "mmap %s\n", strerror(errno));
		return (1);
	}
	iter = p;
	m = (pthread_mutex_t *)((char *)p + sizeof(*iter));
	error = pthread_mutexattr_init(&ma);
	if (error != 0) {
		fprintf(stderr, "pthread_mutexattr_init %s\n",
		    strerror(error));
		return (1);
	}
	error = pthread_mutexattr_setpshared(&ma, 1);
	if (error != 0) {
		fprintf(stderr, "pthread_mutexattr_setpshared %s\n",
		    strerror(error));
		return (1);
	}
	error = pthread_mutex_init(m, &ma);
	if (error != 0) {
		fprintf(stderr, "pthread_mutex_init %s\n",
		    strerror(error));
		return (1);
	}
	error = pthread_mutex_lock(m);
	if (error != 0) {
		fprintf(stderr, "pthread_mutex_lock %s\n",
		    strerror(error));
		return (1);
	}
	child1 = fork();
	if (child1 == -1) {
		fprintf(stderr, "fork %s\n", strerror(errno));
		return (1);
	} else if (child1 == 0) {
		test_mutex(m, &iter->iter1, &iter->aiter1);
		return (0);
	}
	child2 = fork();
	if (child2 == -1) {
		kill(child1, SIGKILL);
		fprintf(stderr, "fork %s\n", strerror(errno));
		return (1);
	} else if (child2 == 0) {
		test_mutex(m, &iter->iter2, &iter->aiter2);
		return (0);
	}
	error = pthread_mutex_unlock(m);
	if (error != 0) {
		fprintf(stderr, "pthread_mutex_unlock %s\n",
		    strerror(error));
		return (1);
	}
	for (i = 0; i < 2;) {
		pid = wait(NULL);
		if (pid == child1 || pid == child2)
			i++;
		else if (pid == -1 && errno != EINTR) {
			fprintf(stderr, "wait %s\n", strerror(errno));
			return (1);
		}
	}
	printf("iter1 %u aiter1 %u iter2 %u aiter2 %u\n",
	    iter->iter1, atomic_load(&iter->aiter1), iter->iter2,
	    atomic_load(&iter->aiter2));
	return (0);
}
